#include <sys/uio.h>
#include <cstring>
#include <cerrno>
#include <sys/mman.h>
#include <fcntl.h>
#include <string>
#include <unistd.h>

#include <trampoline/types.h>

#include "memory.h"
#include "log.h"

// Read Memory
static pid_t guest_pid;
void init_memory(pid_t pid) {
    guest_pid = pid;
}
void memory_writer(uint32_t guest_addr, void *data, uint32_t size) {
    if (size == 0) {
        return;
    }
    iovec local[1];
    local[0].iov_base = data;
    local[0].iov_len = size;
    iovec remote[1];
    remote[0].iov_base = (void *) (uint64_t) guest_addr;
    remote[0].iov_len = size;
    const ssize_t ret = process_vm_writev(guest_pid, local, 1, remote, 1, 0);
    if (ret != size) {
        ERR("Unable To Write Data: %s", strerror(errno));
    }
}

// Shared Memory
static std::string shared_memory_name;
static int shared_memory_fd;
void init_shared_memory_common() {
    shared_memory_name = std::string("/mcpi-trampoline-") + std::to_string(getpid());
    shared_memory_fd = shm_open(shared_memory_name.c_str(), O_CREAT | O_EXCL | O_RDWR, 0600);
    if (shared_memory_fd == -1) {
        ERR("Unable To Create Shared Memory: %s", strerror(errno));
    }
    if (ftruncate(shared_memory_fd, MAX_TRAMPOLINE_ARGS_SIZE) == -1) {
        ERR("Unable To Set Shared Memory Size: %s", strerror(errno));
    }
}
void init_shared_memory_guest() {
    setenv(TRAMPOLINE_SHARED_MEMORY_ENV, shared_memory_name.c_str(), true);
}
unsigned char *shared_memory_host_only;
void init_shared_memory_host() {
    shared_memory_host_only = (unsigned char *) mmap(nullptr, MAX_TRAMPOLINE_ARGS_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, shared_memory_fd, 0);
    if (shared_memory_host_only == MAP_FAILED) {
        ERR("Unable To Map Shared Memory: %s", strerror(errno));
    }
}
void free_shared_memory() {
    munmap(shared_memory_host_only, MAX_TRAMPOLINE_ARGS_SIZE);
    shm_unlink(shared_memory_name.c_str());
}